import re
import common
import os
import sys

from subprocess import Popen, PIPE
from formula import *

from encoders.printers import *


class SyntacticChecker(object):
    
    def __init__(self):
        return


    def checkFalse(self, formula):        

        if formula.nodeType == FNode.CONSTANT and (type(formula.value) == bool):
            return (not formula.value)
        
        if formula.nodeType == FNode.TEXTFORMULA:
            txt = formula.text.strip()
            return txt == "false"

        if formula.nodeType == FNode.NOT:
            return self.checkTrue(formula.formula)

        return False


    def checkTrue(self, formula):        

        if formula.nodeType == FNode.CONSTANT and (type(formula.value) == bool):
            return formula.value
        
        if formula.nodeType == FNode.TEXTFORMULA:
            txt = formula.text.strip()
            return txt == "true"

        if formula.nodeType == FNode.NOT:
            return self.checkFalse(formula.formula)

        return False

    pass


class Z3(object):
    
    def __init__(self, solverpath):
        self.solverpath = solverpath
        return

    
    def createSmtProblem(self, formula):
        f = common.getIOBuffer()

        f.write(common.toBytes("(set-option :print-success false)\n"))
        f.write(common.toBytes("(set-logic LRA)\n"))
        f.write(common.toBytes("\n"))
        
        for v in formula.getDependencies():
            f.write(common.toBytes("(declare-fun %s %s)\n" % (v.name, v.typename.toString())))

        f.write(common.toBytes("\n"))
        f.write(common.toBytes("(assert \n"))

        p = SmtPrinter(f, strictSyntax=True, useBytes=True, smt2version=True)
        p.printer(formula)

        f.write(common.toBytes(")\n"))
        f.write(common.toBytes("(check-sat)\n"))
        f.write(common.toBytes("(exit)\n"))

        data = f.getvalue()
        f.close()

        return data


    def checkSat(self, formula):
        data = self.createSmtProblem(formula)

        output = self.callZ3(data)
        output = common.convertString(output)

        match = re.match("unsat", output)
        return (match is None)


    def callZ3(self, data):
        p = Popen([self.solverpath, "SMTLIB2_COMPLIANT=true", "-smt2", "-in"],
                  stdout=PIPE, stderr=PIPE, stdin=PIPE)
        try:
            res, err = p.communicate(data)
        except KeyboardInterrupt:
            p.kill()
            sys.stdout.write('; Interrupted by Ctrl-C\n')
            exit(1)
            return
        if err:
            raise Exception("ERROR: %s" % err)
        else:
            return res

    pass



class SmtChecker(object):
    
    def __init__(self, solver):
        self.syntactic = SyntacticChecker()
        self.solver = solver
        return


    def checkFalse(self, formula):
        if self.syntactic.checkFalse(formula):
            return True

        return (not self.solver.checkSat(formula))


    def checkTrue(self, formula):
        if self.syntactic.checkFalse(formula):
            return True

        return (not self.solver.checkSat(Not(formula)))

    pass
