from encoders.utils import *
from encoders.helpers import *
from extractors.strategy import *

from extractors.solvers.msat import *
from checkers import *

import itertools
import math

import common
import random


class SimplicesStrategyExtractor(object):

    def __init__(self, environment):
        self.environment = environment
        self.solver = MathSAT5Solver(environment)
        return


    def getLinearStrategy(self, simplex, clausifier):
        yvars = clausifier.yvars
        xvars = clausifier.xvars
        st = self.environment.symbolTable

        s = LinearStrategy(yvars)

        tmpSolver = MathSAT5Solver(self.environment)

        print("      Declaring Variables...")
        coeff = []
        for _ in range(0, len(xvars)):
            row = []
            for _ in range(0, len(yvars) + 1):
                nv = st.createNewFreshSymbol(Types.REAL, "_C_%d")
                row.append(nv)
                tmpSolver.declareVariable(nv)
            coeff.append(row)

        print("      Encoding problem as SMT formula...")
        formula = self.linearEncoding(coeff, simplex, clausifier)
        
        print("      Asserting SMT formula...")
        formula = formula.simplify(True)
        tmpSolver.addAssertion(formula)

        print("      Solving SMT formula...")
        if not tmpSolver.solve():
            return None
                
        print("      Extracting model...")
        for i in range(0, len(xvars)):
            row = coeff[i]
            x = xvars[i]

            vals = []
            for c in row:
                vals.append(tmpSolver.getModelValue(c))
            s.addPolicy(x, vals)

        tmpSolver.destroy()
        return s


    def linearEncoding(self, coeff, simplex, clausifier):
        yvars = clausifier.yvars
        xvars = clausifier.xvars
        ycount = len(yvars)

        print("        Preparing formula definitions...")
        formula = clausifier.getPsi()

        print("        Building problem formula...")
        for i in range(0, len(xvars)):
            row = coeff[i]
            x = xvars[i]

            f = row[-1]
            for j in range(0, len(row) - 1):
                c = row[j]
                f = Plus(Times(c, yvars[j]), f)

            formula = Definition(x.name, f, formula)

        formula = formula.simplify()

        print("        Creating simplex encoding...")
        reslist = []
        for point in simplex:
            start = formula
            for i in range(0, len(point)):
                start = Definition(yvars[i].name, Constant(point[i]), start)
            reslist.append(start)

        return And(reslist)


    def extractStrategy(self, clausifier):
        yvars = clausifier.yvars

        print("Starting strategy extraction...")

        s = PiecewiseStrategy(yvars)

        ran = range(0, len(yvars))
        c = 0
        tot = math.factorial(len(ran))
        for p in itertools.permutations(ran):
            
            c = c + 1
            perc = float(c * 100) / float(tot)
            print("\n  [%.2f%%] Examining region %d over %d..." % (perc, c, tot))
            print("    Generating new simplex...")
            simplex = []
            for i in range(0, len(yvars)+1):
                row = [0 for _ in range(0, len(yvars))]
                for j in range(0, len(yvars)):
                    idx = p[j]
                    if j >= i:
                        (_, u) = clausifier.getYBound(yvars[idx])
                        row[idx] = u
                simplex.append(row)

            print("    Generating simplex symbolic region...")
            regionClauses = []
            for i in range(1, len(yvars)):
                r = LE(yvars[p[i-1]], yvars[p[i]])
                regionClauses.append(r)
            region = And(regionClauses)
            
            print("    Generating a linear startegy for the simplex...")
            p = self.getLinearStrategy(simplex, clausifier)
            assert p is not None
            
            print("    Adding clause to global policy...")
            s.addPiece(region, p)

        return s
    pass
