from encoders.printers import *
from formula import *
import sys


class LinearStrategy(object):

    def __init__(self, yvars):
        self.ycount = len(yvars)
        self.yvars = list(yvars)

        self.matrix = {}
        return


    def addPolicy(self, xvar, coefficients):
        assert len(coefficients) == self.ycount + 1

        self.matrix[xvar] = list(coefficients)

        return


    def addZeroPolicy(self, xvar):
        coefficients = []
        for _ in range(0, self.ycount + 1):
            coefficients.append((0,1))

        self.matrix[xvar] = list(coefficients)

        return


    def getXVars(self):
        return list(self.matrix.keys())


    def getPolicyAsFormula(self, xvar):
        assert xvar in self.matrix

        coeff = self.matrix[xvar]

        f = Constant(coeff[-1])

        for i in range(0, self.ycount):
            frac = coeff[i]
            if frac != 0:
                f = Plus(Times(Constant(frac), self.yvars[i]), f)

        return f


    def getStrategyAsFormula(self):
        res = []
        for k in self.matrix.keys():
            res.append(self.getPolicyAsFormula(k))
        return And(res)


    def dump(self, stream=sys.stdout):

        for xvar, coeff in self.matrix.items():
            stream.write(xvar.name + " = ")
            for i in range(0, self.ycount):
                f = coeff[i]
                if f != 0:
                    stream.write("(" + self.printCoeff(f))
                    stream.write(" * " + self.yvars[i].name + ")")
                    stream.write(" + ")

            stream.write(self.printCoeff(coeff[-1]))
            stream.write("\n")
        return


    def printCoeff(self, c):
        return str(c)
        # res = ""
        # (n,d) = c
        # res = str(n)
        # if d != 1:
        #     res = res + "/" + str(d)
        # return res

    pass



class PiecewiseStrategy(object):

    def __init__(self, yvars):
        self.yvars = list(yvars)

        self.pieces = []
        return

    def printFormula(self, formula, stream=sys.stdout):
        p = HRPrinter(stream, useBytes=(stream!=sys.stdout))
        p.printer(formula)
        return

    def addPiece(self, region, strategy):
        assert len(strategy.yvars) == len(self.yvars)

        if isinstance(strategy, PiecewiseStrategy):
            for (r,s) in strategy.getPieces():
                tmp = And([region, r])
                self.addPiece(tmp, s)

        else:
            self.pieces.append((region, strategy))

        return


    def getPieces(self):
        return self.pieces


    def dump(self, stream=sys.stdout):
        assert len(self.pieces) > 0

        if len(self.pieces) == 1:
            for (region, strategy) in self.pieces:
                strategy.dump()
                return

        count = 0
        for (region, strategy) in self.pieces:
            if count == 0:
                stream.write("IF (")
                count = 1
            else:
                stream.write("ELSE IF (")

            self.printFormula(region, stream)
            stream.write(") THEN\n")
            strategy.dump(stream)
            stream.write('\n')
        return


    pass
